package eu.asvachin.hashcalculator;

import eu.asvachin.hashcalculator.hashes.Hash;
import eu.asvachin.hashcalculator.hashes.Hash.HashType;
import eu.asvachin.hashcalculator.hashes.crc.CRC;
import eu.asvachin.hashcalculator.hashes.crc.CRC32Adler;
import eu.asvachin.hashcalculator.hashes.md.MD2;
import eu.asvachin.hashcalculator.hashes.md.MD4;
import eu.asvachin.hashcalculator.hashes.md.MD5;
import eu.asvachin.hashcalculator.hashes.ripemd.RIPEMD128;
import eu.asvachin.hashcalculator.hashes.ripemd.RIPEMD160;
import eu.asvachin.hashcalculator.hashes.ripemd.RIPEMD256;
import eu.asvachin.hashcalculator.hashes.ripemd.RIPEMD320;
import eu.asvachin.hashcalculator.hashes.sha.*;
import eu.asvachin.hashcalculator.hashes.sum.Sum;
import eu.asvachin.hashcalculator.hashes.sum.Xor8;
import eu.asvachin.hashcalculator.hashes.tiger.Tiger;
import eu.asvachin.hashcalculator.hashes.whirlpool.Whirlpool;
import eu.asvachin.toolbox.properties.Properties;
import java.io.IOException;
import java.io.InputStream;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Reads a file and sends the data to the threads that do the actual hashing.
 *
 * @author Miles Asvachin
 */
public final class HashController
{
	/** All hashes that are supported. */
	private final List<Hash> hashes = new ArrayList<>();
	/** Hashes currently being used. */
	private final Collection<Hasher> hashers = new LinkedList<>();
	/** Queue of hashes waiting for data. */
	private final Deque<Hasher> waitingHashers = new LinkedList<>();
	private volatile boolean doHashing;
	private volatile boolean pauseHashing;
	
	public HashController()
	{
		//Checksums:
		hashes.add(new Sum(8));
		hashes.add(new Sum(16));
		hashes.add(new Sum(24));
		hashes.add(new Sum(32));
		hashes.add(new Sum(64));
		hashes.add(new Xor8());
		
		//CRCs:
		hashes.add(CRC.createCRC("CRC-5-CCITT", 5, 0x15, 0x0, false, false, 0x0));
		hashes.add(CRC.createCRC("CRC-5-USB", 5, 0x05, 0x0, false, false, 0x0));
		hashes.add(CRC.createCRC("CRC-7", 7, 0x09, 0x0, false, false, 0x0));
		hashes.add(CRC.createCRC("CRC-8", 8, 0xEA, 0x0, false, false, 0x0));
		hashes.add(CRC.createCRC("CRC-8-ATM", 8, 0x07, 0x0, false, false, 0x0));
		hashes.add(CRC.createCRC("CRC-10", 10, 0x233, 0x0, false, false, 0x0));
		hashes.add(CRC.createCRC("CRC-12", 12, 0x80F, 0x0, false, false, 0x0));
		hashes.add(CRC.createCRC("CRC-16-BBS", 16, 0x8408, 0x0, true, true, 0x0));
		hashes.add(CRC.createCRC("CRC-16-CCITT", 16, 0x1021, 0xFFFF, false, false, 0x0));
		hashes.add(CRC.createCRC("CRC-16-IBM", 16, 0x8005, 0x0, true, true, 0x0));
		hashes.add(CRC.createCRC("CRC-24", 24, 0x864CFB, 0xB704CE, false, false, 0x0));
		hashes.add(new CRC32Adler());
		hashes.add(CRC.createCRC("CRC-32-C", 32, 0x1EDC6F41, 0xFFFFFFFF, true, true, 0xFFFFFFFF));
		hashes.add(CRC.createCRC("CRC-32-IEEE 802.3", 32, 0x04C11DB7, 0xFFFFFFFF, true, true, 0xFFFFFFFF));
		
		//Cryptographic Hashes:
		hashes.add(new MD2());
		hashes.add(new MD4());
		hashes.add(new MD5());
		hashes.add(new RIPEMD128());
		hashes.add(new RIPEMD160());
		hashes.add(new RIPEMD256());
		hashes.add(new RIPEMD320());
		hashes.add(new SHA1());
		hashes.add(new SHA224());
		hashes.add(new SHA256());
		hashes.add(new SHA384());
		hashes.add(new SHA512());
		hashes.add(new Tiger(128));
		hashes.add(new Tiger(160));
		hashes.add(new Tiger());
		hashes.add(new Whirlpool());
	}
	
	/**
	 * @return the names of all supported hash functions, grouped by type.
	 */
	public Map<HashType, List<String>> getHashNames()
	{
		final Map<HashType, List<String>> hashNames = new EnumMap<>(HashType.class);
		
		for(final HashType hashType : HashType.values())
		{
			hashNames.put(hashType, new ArrayList<String>());
		}

		for(final Hash hash : hashes)
		{
			hashNames.get(hash.getHashType()).add(hash.getHashName());
		}
		
		return hashNames;
	}
	
	/**
	 * Temporarily stop hashing. May take a short while to stop while buffered data is processed.
	 */
	public synchronized void pauseHashing()
	{
		if(doHashing)
		{
			pauseHashing = true;
		}
	}
	
	/**
	 * Resume after pausing.
	 */
	public synchronized void resumeHashing()
	{
		if(pauseHashing)
		{
			pauseHashing = false;
			notifyAll();
		}
	}
	
	/**
	 * Stop hashing, discarding any results calculated so far. May take a short while to stop while buffered data is processed.
	 */
	public synchronized void cancelHashing()
	{
		if(doHashing)
		{
			doHashing = false;

			if(pauseHashing)
			{
				resumeHashing();
			}
		}
	}
	
	/**
	 * @return true if hashes are in the process of being calculated, even if paused.
	 */
	public boolean isHashing()
	{
		return doHashing;
	}
	
	/**
	 * @return true if hashes are in the process of being calculated and also paused.
	 */
	public boolean isPaused()
	{
		return doHashing && pauseHashing;
	}
	
	/**
	 * Calculate the hashes of the given data.
	 * 
	 * @param hashesToUse the names of the hash functions to use.
	 * @param dataStream the data to hash. The calling code needs to take care of closing the stream.
	 * @param progressListener object wanting to be notified of progress and time taken to complete.
	 * @return the calculated hashes grouped by hash type and indexed by hash name.
	 */
	public Map<HashType, Map<String, String>> hash(final Collection<String> hashesToUse, final InputStream dataStream, final HashProgressListener progressListener) throws IOException, InterruptedException
	{
		Logger.getLogger(getClass().getName()).log(Level.INFO, "Started hashing.");
		
		doHashing = true;
		pauseHashing = false;
		createThreads(hashesToUse);
		waitingHashers.clear();
		waitingHashers.addAll(hashers);

		final int bufferSize = Properties.getAppProperties().getProperty(int.class, "BufferSize");
		final byte[] buffer = new byte[bufferSize];
		int bytesRead;
		long totalBytesRead = 0;
		final long startTime = System.currentTimeMillis();

		//n.b. Reading into a buffer of bytes is significantly faster than a
		//BufferedInputStream, and many times faster than an unbuffered FileInputStream.
		while(doHashing)
		{
			progressListener.setProgress(totalBytesRead);
			bytesRead = dataStream.read(buffer);

			if(bytesRead > 0)
			{
				updateWaitingHashers(buffer, bytesRead);
				totalBytesRead += bytesRead;
			}
			else if(bytesRead < 0)
			{
				break;
			}
			
			if(pauseHashing)
			{
				synchronized(this)
				{
					progressListener.paused();

					while(pauseHashing)
					{
						wait();
					}
				
					progressListener.resumed();
				}
			}
		}

		finishWaitingHashers();
		hashers.clear();

		final Map<HashType, Map<String, String>> results;
		
		if(doHashing)
		{
			final long finishTime = System.currentTimeMillis();
			final long timeTaken = finishTime - startTime;
			progressListener.completed(timeTaken);

			results = new EnumMap<>(HashType.class);

			for(final HashType hashType : HashType.values())
			{
				results.put(hashType, new HashMap<String, String>());
			}

			for(final String hashName : hashesToUse)
			{
				for(final Hash hash : hashes)
				{
					if(hash.getHashName().equals(hashName))
					{
						results.get(hash.getHashType()).put(hashName, hash.getDigest());
					}
				}
			}

			Logger.getLogger(getClass().getName()).log(Level.INFO, "Finished hashing {0}bytes in {1}ms.", new Object[]{totalBytesRead, timeTaken});
		}
		else
		{
			results = null;
			progressListener.cancelled();
		}
		
		return results;
	}
	
	/**
	 * Spawns a bunch of threads.
	 * 
	 * @param hashesToUse one thread will be created for each of the named hashes.
	 */
	private void createThreads(final Collection<String> hashesToUse)
	{
		for(final String hashName : hashesToUse)
		{
			for(final Hash hash : hashes)
			{
				if(hash.getHashName().equals(hashName))
				{
					final Hasher hasher = new Hasher(hash, new HashQueue()
					{
						@Override
						public void addWaitingHasher(Hasher waitingHasher)
						{
							synchronized(HashController.this)
							{
								//Use LIFO so that the slowest hashes get sent new data first, slightly improving parallelism.
								waitingHashers.push(waitingHasher);
								HashController.this.notifyAll();
							}
						}
					});
					hashers.add(hasher);

					final Thread thread = new Thread(hasher, hash.getHashName());
					thread.start();
				}
			}
		}
	}
	
	/**
	 * Waits until all hashers have finished processing their current data, then sends them the new data, emptying the waitingHashers queue in the process.
	 * 
	 * @param newData
	 * @param length how many bytes in newData are actually data.
	 */
	private synchronized void updateWaitingHashers(final byte[] newData, final int length)
	{
		while(waitingHashers.size() < hashers.size())
		{
			try
			{
				wait();
			}
			catch(InterruptedException ex){}
		}

		while(!waitingHashers.isEmpty())
		{
			waitingHashers.pop().update(newData, length);
		}
	}

	/**
	 * Waits until all hashers have finished processing, then tells them to finish, emptying the waitingHashers queue in the process.
	 */
	private synchronized void finishWaitingHashers()
	{
		while(waitingHashers.size() < hashers.size())
		{
			try
			{
				wait();
			}
			catch(InterruptedException ex){}
		}
		
		while(!waitingHashers.isEmpty())
		{
			waitingHashers.pop().finish();
		}
	}
	
	public interface HashQueue
	{
		void addWaitingHasher(final Hasher waitingHasher);
	}
	
	/**
	 * Receives notifications about the current state of the hashing process.
	 */
	public interface HashProgressListener
	{
		void setProgress(final long bytesHashed);
		void completed(final long milliseconds);
		void paused();
		void resumed();
		void cancelled();
	}
}
