package eu.asvachin.hashcalculator.gui;

import eu.asvachin.hashcalculator.HashController;
import eu.asvachin.hashcalculator.HashController.HashProgressListener;
import eu.asvachin.hashcalculator.hashes.Hash;
import eu.asvachin.hashcalculator.hashes.Hash.HashType;
import eu.asvachin.hashcalculator.utilities.StringUtilities;
import eu.asvachin.toolbox.properties.Properties;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.util.*;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumn;

/**
 *
 * @author Miles Asvachin
 */
public final class HashCalculatorGui implements HashProgressListener
{
	/** Calculates the hashes. */
	private final HashController controller;
	/** These hold the results. */
	private final Map<HashType, HashTableModel> tableModels;
	private final JProgressBar progressBar = new JProgressBar(0, 100);
	/** Path to current file. */
	final JTextField filePathText = new JTextField(32);
	/** This is what will be hashed when the start button is pressed. */
	private File currentFile;
	
	public HashCalculatorGui(final HashController newController)
	{
		controller = newController;
		
		try
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		}
		catch(ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException ex)
		{
			Logger.getLogger(getClass().getName()).log(Level.WARNING, null, ex);
		}
		
		JFrame.setDefaultLookAndFeelDecorated(true);
		final JFrame frame = new JFrame(Properties.getToolboxProperties().getProperty(String.class, "AppName"));
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		//frame.setIconImage(new ImageIcon(getClass().getResource("Icon32.png"), "Icon32").getImage());
		
		final JComponent contentPane = new Box(BoxLayout.Y_AXIS);
		frame.setContentPane(contentPane);
		
		contentPane.add(createTopPanel());
		tableModels = new EnumMap<>(HashType.class);
		contentPane.add(createHashTables(controller.getHashNames()));
		
		progressBar.setValue(0);
		progressBar.setStringPainted(true);
		progressBar.setString("Waiting");
		contentPane.add(progressBar);
		
		new FileDrop(frame, new FileDrop.Listener()
		{
			@Override
			public void filesDropped(final File[] files)
			{
				filePathText.setText(files[0].getAbsolutePath());
				filePathText.getInputVerifier().verify(filePathText);
			}
		});
		
		//frame.setResizable(false);
		frame.pack();
		frame.setMaximumSize(new Dimension(Integer.MAX_VALUE, frame.getPreferredSize().height));
		frame.setVisible(true);
	}
	
	/**
	 * @return controls for selecting the file and starting/stopping hashing.
	 */
	private JComponent createTopPanel()
	{
		final JComponent container = new Box(BoxLayout.X_AXIS);
		
		final JButton fileButton = new JButton(new ImageIcon(getClass().getResource("Open24.png")));
		fileButton.setToolTipText("Open file");
		fileButton.requestFocusInWindow();
		container.add(fileButton);
		
		final JComponent fileDetailsContainer = new Box(BoxLayout.Y_AXIS);
		fileDetailsContainer.add(filePathText);
		final JLabel fileSizeText = new JLabel("-");
		fileDetailsContainer.add(fileSizeText);
		container.add(fileDetailsContainer);
		
		final JComponent controlsContainer = new Box(BoxLayout.X_AXIS);
		final JButton startButton = new JButton(new ImageIcon(getClass().getResource("Play24.png")));
		startButton.setToolTipText("Start hashing");
		controlsContainer.add(startButton);
		final JButton pauseButton = new JButton(new ImageIcon(getClass().getResource("Pause24.png")));
		pauseButton.setToolTipText("Pause hashing");
		pauseButton.setEnabled(false);
		controlsContainer.add(pauseButton);
		final JButton stopButton = new JButton(new ImageIcon(getClass().getResource("Stop24.png")));
		stopButton.setToolTipText("Stop hashing");
		stopButton.setEnabled(false);
		controlsContainer.add(stopButton);
		container.add(controlsContainer);
		container.setMaximumSize(new Dimension(Integer.MAX_VALUE, container.getPreferredSize().height));
		
		final FileInputVerifier verifier = new FileInputVerifier()
		{
			@Override
			public void processFileDetails(final File file)
			{
				if(file != null && file.exists())
				{
					fileSizeText.setText(file.length() + " bytes");
				}
			}

			@Override
			public boolean verify(final JComponent input)
			{
				currentFile = new File(filePathText.getText());
				processFileDetails(currentFile);
				
				return true; //Dont' prevent focus being lost.
			}

			@Override
			public void actionPerformed(ActionEvent e)
			{
				filePathText.transferFocus();
			}
		};
		
		filePathText.addActionListener(verifier);
		filePathText.setInputVerifier(verifier);
		
		fileButton.addActionListener(new ActionListener()
		{
			private final JFileChooser fileChooser = new JFileChooser();

			@Override
			public void actionPerformed(ActionEvent e)
			{
				if(currentFile != null)
				{
					fileChooser.setCurrentDirectory(currentFile.getParentFile());
				}

				if(fileChooser.showOpenDialog(container) == JFileChooser.APPROVE_OPTION)
				{
					currentFile = fileChooser.getSelectedFile();
					filePathText.setText(currentFile.getAbsolutePath());
					verifier.processFileDetails(currentFile);
					
					for(final Entry<HashType, HashTableModel> entry : tableModels.entrySet())
					{
						entry.getValue().setFileHashes(null);
						progressBar.setValue(0);
						progressBar.setString("Waiting");
					}
				}
			}
		});
		
		startButton.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				if(controller.isPaused())
				{
					controller.resumeHashing();
				}
				else
				{
					if(currentFile != null && currentFile.exists())
					{
						startButton.setEnabled(false);
						pauseButton.setEnabled(true);
						stopButton.setEnabled(true);
					
						final Map<Hash.HashType, List<String>> allHashes = controller.getHashNames();						
						final Collection<String> enabledHashNames = new ArrayList<>(allHashes.size());
						
						for(final Entry<HashType, HashTableModel> entry : tableModels.entrySet())
						{
							enabledHashNames.addAll(entry.getValue().getEnabledHashNames());
							entry.getValue().setFileHashes(null);
						}
						
						saveEnabledHashNames(enabledHashNames);

						final Thread thread = new Thread(new Runnable()
						{
							@Override
							public void run()
							{
								try(final InputStream fileIS = new FileInputStream(currentFile))
								{
									final Map<HashType, Map<String, String>> results = controller.hash(enabledHashNames, fileIS, HashCalculatorGui.this);

									if(results != null)
									{
										for(final Entry<HashType, Map<String, String>> entry : results.entrySet())
										{
											tableModels.get(entry.getKey()).setFileHashes(entry.getValue());
										}
									}
								}
								catch(InterruptedException | IOException ex)
								{
									//@todo Show error message.
								}
								finally
								{
									startButton.setEnabled(true);
									pauseButton.setEnabled(false);
									stopButton.setEnabled(false);
								}
							}
						});
						thread.start();
					}
					else
					{
						//@todo Show error message.
					}
				}
			}
		});
		
		pauseButton.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				if(controller.isPaused())
				{
					controller.resumeHashing();
					startButton.setEnabled(false);
				}
				else
				{
					controller.pauseHashing();
					startButton.setEnabled(true);
				}
			}
		});
		
		stopButton.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				controller.cancelHashing();
				startButton.setEnabled(true);
				pauseButton.setEnabled(false);
				stopButton.setEnabled(false);
			}
		});
		
		return container;
	}
	
	/**
	 * @param hashNames names of hashes that will be listed in the tables.
	 * @return a table per type of hash.
	 */
	private JComponent createHashTables(final Map<Hash.HashType, List<String>> hashNames)
	{
		final Collection<JTable> tables = new ArrayList<>(hashNames.size());
		final JComponent container = new Box(BoxLayout.Y_AXIS);
		final Collection<String> enabledHashNames = getEnabledHashNames();

		for(final Entry<Hash.HashType, List<String>> entry : hashNames.entrySet())
		{
			final HashTableModel tableModel = new HashTableModel(entry.getValue(), enabledHashNames);
			tableModels.put(entry.getKey(), tableModel);
			
			final JComponent tableContainer = new JPanel(new BorderLayout());
			final JTable table = new JTable(tableModel);
			tables.add(table);
			table.setDefaultRenderer(String.class, new HashTableCellRenderer());
			table.setDefaultRenderer(BigInteger.class, new HashTableCellRenderer());
			table.setDefaultRenderer(Boolean.class, new TickTableCellRenderer());
			table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			table.setCellSelectionEnabled(true);
			//Default behaviour is to copy a whole row, even if cell selection is turned on. Changed to copying/pasting single cells.
			table.setTransferHandler(new TransferHandler()
			{
				@Override
				public boolean canImport(final JComponent comp, final DataFlavor[] transferFlavors)
				{
					boolean result = false;
					
					for(final DataFlavor flavor : transferFlavors)
					{
						if(flavor.equals(DataFlavor.stringFlavor))
						{
							result = true;
							break;
						}
					}
					
					return result;
				}
				
				@Override
				public boolean importData(final JComponent c, final Transferable t)
				{
					boolean imported = false;
					final int columnIndex = table.getSelectedColumn();
					final int rowIndex = table.getSelectedRow();
					
					if(columnIndex > -1 && rowIndex > -1
							&& table.isCellEditable(rowIndex, columnIndex)
							&& canImport(c, t.getTransferDataFlavors()))
					{
						try
						{
							final String value = (String)t.getTransferData(DataFlavor.stringFlavor);
							table.setValueAt(value, rowIndex, columnIndex);
							imported = true;
						}
						catch(UnsupportedFlavorException | IOException ex)
						{
							Logger.getLogger(getClass().getName()).log(Level.WARNING, "Failed to paste string from clipboard.", ex);
						}
					}

					return imported;
				}
				
				@Override
				public int getSourceActions(final JComponent c)
				{
					return COPY;
				}
				
				@Override
				protected Transferable createTransferable(JComponent c)
				{
					final int columnIndex = table.getSelectedColumn();
					final int rowIndex = table.getSelectedRow();
					final Transferable trans;
					
					if(columnIndex > -1 && rowIndex > -1)
					{
						final Object value = table.getValueAt(rowIndex, columnIndex);
						trans = value == null ? null : new StringSelection(value.toString());
					}
					else
					{
						trans = null;
					}
					
					return trans;
				}
			});

			//Table headers are not shown by default unless the table is in a JScrollPane. They don't
			//need to scroll individually, so show the header separately instead.
			final JTableHeader header = table.getTableHeader();
			header.setReorderingAllowed(false);
			
			final JLabel title = new JLabel(entry.getKey().toString());
			title.setBorder(BorderFactory.createRaisedBevelBorder());
			title.addMouseListener(new MouseAdapter()
			{
				@Override
				public void mouseClicked(final MouseEvent e)
				{
					if(e.getClickCount() == 2)
					{
						((HashTableModel)table.getModel()).toggleAllHashesEnabled();
					}
				}
			});
			
			tableContainer.add(title, BorderLayout.NORTH);
			tableContainer.add(header, BorderLayout.CENTER);
			tableContainer.add(table, BorderLayout.SOUTH);
			container.add(tableContainer);
		}
		
		for(final JTable table : tables)
		{
			for(final TableColumn column : Collections.list(table.getColumnModel().getColumns()))
			{
				column.setResizable(false);
				column.setHeaderRenderer(new HashTableHeaderRenderer(tables));
			}
			
			table.addFocusListener(new FocusListener()
			{
				@Override
				public void focusGained(final FocusEvent e)
				{
					for(final JTable otherTable : tables)
					{
						if(otherTable != table)
						{
							otherTable.clearSelection();
						}
					}
				}

				@Override
				public void focusLost(final FocusEvent e){}
			});
			
			table.addMouseListener(new MouseAdapter()
			{
				@Override
				public void mouseClicked(final MouseEvent e)
				{
					if(e.getClickCount() == 2)
					{
						final Point point = e.getPoint();
						final int rowIndex = table.rowAtPoint(point);

						((HashTableModel)table.getModel()).toggleHashEnabled(rowIndex);
					}
				}
			});
		}
		
		return new JScrollPane(container);
	}
	
	/**
	 * @return previously saved list of enabled hashes.
	 */
	private static Collection<String> getEnabledHashNames()
	{
		final String rawString = Properties.getAppProperties().getProperty(String.class, "EnabledHashes");
		final String[] hashNameArray = rawString.split("[\r?\n]+");
		
		return Arrays.asList(hashNameArray);
	}
	
	/**
	 * Saves a list of the currently enabled hashes.
	 */
	private static void saveEnabledHashNames(final Collection<String> enabledHashNames)
	{
		final StringBuilder joined = new StringBuilder();
		
		for(final String hashName : enabledHashNames)
		{
			joined.append(hashName);
			joined.append("\n"); //Use only "\n" because "\r" is saved as the character sequence "&#13;" instead of just the character itself.
		}
		
		Properties.getAppProperties().setProperty(String.class, joined.toString(), "EnabledHashes");
		Properties.getAppProperties().saveProperties();
	}

	@Override
	public void setProgress(final long bytesHashed)
	{
		final int progress = (int)(((double)bytesHashed / (double)currentFile.length()) * 100.0);
		progressBar.setValue(progress);
		progressBar.setString(progress + "%");
	}

	@Override
	public void completed(final long milliseconds)
	{
		progressBar.setString("Completed in " + StringUtilities.formatTime(milliseconds));
	}

	@Override
	public void paused()
	{
		progressBar.setString("Paused");
	}

	@Override
	public void resumed()
	{
		progressBar.setString("Resuming...");
	}

	@Override
	public void cancelled()
	{
		progressBar.setString("Cancelled");
	}
	
	private abstract class FileInputVerifier extends InputVerifier implements ActionListener
	{
		public abstract void processFileDetails(final File file);
	}
}
