package eu.asvachin.hashcalculator.gui;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import javax.swing.table.AbstractTableModel;

/**
 * Stores hash results.
 *
 * @author Miles Asvachin
 */
public final class HashTableModel extends AbstractTableModel
{
	enum ColumnDescription
	{
		HASH_NAME("Hash Name", String.class),
		FILE_HASH("File Hash", String.class),
		COMP_HASH("Comparison Hash", String.class),
		MATCH("", Boolean.class);
		
		public final String name;
		public final Class columnClass;

		ColumnDescription(final String newName, Class newClass)
		{
			name = newName;
			columnClass = newClass;
		}
	}
	
	final private List<HashRow> rows;
	
	/**
	 * @param hashNames all hash names that will be shown in the table.
	 * @param enabledHashNames if an element of hashNames is also in here, it will be enabled, otherwise it will be disabled.
	 */
	public HashTableModel(final List<String> hashNames, final Collection<String> enabledHashNames)
	{
		rows = new ArrayList<>(hashNames.size());
		
		for(final String hashName : hashNames)
		{
			final HashRow row = new HashRow(hashName);
			row.enabled = enabledHashNames.contains(hashName);
			rows.add(row);
		}
	}
	
	/**
	 * Store some new results.
	 */
	public void setFileHashes(final Map<String, String> newHashes)
	{
		for(final HashRow row : rows)
		{
			row.fileHash = newHashes == null ? null : newHashes.get(row.hashName);
		}
		
		fireTableDataChanged();
	}
	
	/**
	 * toggles the enabled status of a row.
	 */
	public void toggleHashEnabled(final int rowIndex)
	{
		final HashRow row = rows.get(rowIndex);
		row.enabled = !row.enabled;
		this.fireTableRowsUpdated(rowIndex, rowIndex);
	}
	
	/**
	 * if any rows are currently enabled, this disables all rows. Otherwise it enables all rows.
	 */
	public void toggleAllHashesEnabled()
	{
		boolean anyHashesEnabled = false;
		
		for(final HashRow row : rows)
		{
			if(row.enabled)
			{
				anyHashesEnabled = true;
				break;
			}
		}
		
		for(final HashRow row : rows)
		{
			row.enabled = !anyHashesEnabled;
		}
		
		fireTableDataChanged();
	}
	
	public boolean isHashEnabled(final int rowIndex)
	{
		return rows.get(rowIndex).enabled;
	}
	
	/**
	 * @return a list of hash names for currently enabled rows.
	 */
	public Collection<String> getEnabledHashNames()
	{
		final Collection<String> enabledHashNames = new ArrayList<>(rows.size());

		for(final HashRow row : rows)
		{
			if(row.enabled)
			{
				enabledHashNames.add(row.hashName);
			}
		}
		
		return enabledHashNames;
	}
	
	@Override
	public void setValueAt(final Object aValue, final int rowIndex, final int columnIndex)
	{
		if(ColumnDescription.values()[columnIndex] == ColumnDescription.COMP_HASH)
		{
			if(aValue == null || aValue.toString().isEmpty())
			{
				rows.get(rowIndex).comparisonHash = "";
			}
			else
			{
				try
				{
					new BigInteger(aValue.toString(), 16); //This is just to cause an exception if the string is not a hex number.
					rows.get(rowIndex).comparisonHash = aValue.toString().toUpperCase();
				}
				catch(NumberFormatException ex)
				{
					//@todo Highlight the error somehow.
				}
			}
			
			fireTableDataChanged();
		}
    }
	
	@Override
	public String getColumnName(final int columnIndex)
	{
		return ColumnDescription.values()[columnIndex].name;
	}
	
	@Override
	public Class<?> getColumnClass(final int columnIndex)
	{
        return ColumnDescription.values()[columnIndex].columnClass;
    }
	
	@Override
	public int getRowCount()
	{
		return rows.size();
	}

	@Override
	public int getColumnCount()
	{
		return ColumnDescription.values().length;
	}

	@Override
	public Object getValueAt(final int rowIndex, final int columnIndex)
	{
		switch(columnIndex)
		{
			case 0:
				return rows.get(rowIndex).hashName;
			case 1:
				return rows.get(rowIndex).fileHash;
			case 2:
				return rows.get(rowIndex).comparisonHash;
			case 3:
				final String fileHash = rows.get(rowIndex).fileHash;
				final String comparisonHash = rows.get(rowIndex).comparisonHash;
				return comparisonHash == null || comparisonHash.isEmpty() || fileHash == null || fileHash.isEmpty()
						? null
						: comparisonHash.equalsIgnoreCase(fileHash);
			default:
				return null;
		}
	}
	
	@Override
	public boolean isCellEditable(final int rowIndex, final int columnIndex)
	{
        return ColumnDescription.values()[columnIndex] == ColumnDescription.COMP_HASH;
    }
	
	private static class HashRow
	{
		final String hashName;
		String fileHash;
		String comparisonHash;
		boolean enabled = false;
		
		HashRow(final String newHashName)
		{
			hashName = newHashName;
		}
	}
}
