package eu.asvachin.hashcalculator.hashes.crc;

import eu.asvachin.hashcalculator.utilities.BitUtilities;
import java.math.BigInteger;

/**
 * Same as LongCRC but without the upper limit to width.
 * Ok, so it's still limited, but to a ridiculous 2^31 bits wide.
 * WARNING: Really, really, slow relative to smaller width versions!
 *
 * @version 1.0 20 June 2006
 * @author Miles Asvachin
 */
public class BigCRC extends CRC
{
	//CRC variables:
	private final int width; //Width of the algorithm.
	private final BigInteger polynomial; //Polynomial representation.
	private final BigInteger initialValue; //Initial value of shiftRegister.
	private final boolean reflectIn; //Reverse order of input bits.
	private final boolean reflectOut; //Reverse order of output bits.
	private final BigInteger finalXORValue; //XOR the output with this.
	
	//Variables used for calculation:
	private BigInteger shiftRegister;
	private BigInteger[] lookupTable = new BigInteger[256];
	private int minBytes;
	
	protected BigCRC(String name, int w, BigInteger poly, BigInteger init, boolean refIn, boolean refOut, BigInteger xorVal)
	{
		hashName = name;
		width = w;
		initialValue = init;
		reflectIn = refIn;
		reflectOut = refIn ^ refOut; //Account for reflected algorithm already reflecting output.
		finalXORValue = xorVal;
		polynomial = refIn ? BitUtilities.reflect(poly, width) : poly;
		minBytes = (int)Math.ceil(width / 8.0);
		shiftRegister = init;
		calcTable();
	}
	
	@Override
	public void reset(){shiftRegister = new BigInteger(initialValue.toString());}
	
	/**
	 * Precalculate values to XOR with register for every possible input byte.
	 */
	private void calcTable()
	{
		if(reflectIn)
		{
			for(int i = 0; i < lookupTable.length; i++)
			{
				lookupTable[i] = new BigInteger(Integer.toString(i)) ;
				
				for(int j = 0; j < 8; j++)
				{
					if(lookupTable[i].testBit(0))
					{
						lookupTable[i] = BitUtilities.truncate(lookupTable[i].shiftRight(1), minBytes).xor(polynomial);
					}
					else
					{
						lookupTable[i] = BitUtilities.truncate(lookupTable[i].shiftRight(1), minBytes);
					}
				}
			}
		}
		else
		{
			for(int i = 0; i < lookupTable.length; i++)
			{
				lookupTable[i] = (new BigInteger(Integer.toString(i))).shiftLeft(width - 8);
				
				for(int j = 0; j < 8; j++)
				{
					if(lookupTable[i].testBit(width - 1))
					{
						lookupTable[i] = BitUtilities.truncate(lookupTable[i].shiftLeft(1), minBytes).xor(polynomial);
					}
					else
					{
						lookupTable[i] = BitUtilities.truncate(lookupTable[i].shiftLeft(1), minBytes);
					}
				}
			}
		}
	}
	
	@Override
	public void update(final byte[] newData)
	{
		for(final byte b : newData)
		{
			if(reflectIn)
			{
				shiftRegister = lookupTable[(shiftRegister.intValue() ^ b) & 0xFF].xor(shiftRegister.shiftRight(8));
			}
			else
			{
				shiftRegister = lookupTable[(shiftRegister.shiftRight(width - 8).intValue() ^ b) & 0xFF].xor(shiftRegister.shiftLeft(8));
			}

			//truncate is slow, so don't do it too often.
			if(shiftRegister.bitLength() > width + 24)
			{
				shiftRegister = BitUtilities.truncate(shiftRegister, minBytes);
			}
		}
	}
	
	@Override
	public void finish()
	{
		//Reverse output bit order.
		if(reflectOut)
		{
			shiftRegister = BitUtilities.reflect(shiftRegister, width);
		}
		
		if(finalXORValue.bitLength() > 0)
		{
			shiftRegister = shiftRegister.xor(finalXORValue);
		}
		
		String digestFormat = "%" + "0" + (minBytes * 2) + "x";
		BigInteger hashMask = BigInteger.ONE;
		
		for(int i = 1; i < width; i++)
		{
			hashMask = hashMask.shiftLeft(1).setBit(0);
		}
		
		digest = String.format(digestFormat, shiftRegister.and(hashMask));
		reset();
	}
}