package eu.asvachin.hashcalculator.hashes.crc;

import eu.asvachin.hashcalculator.hashes.Hash;
import java.math.BigInteger;

/**
 * Base class for all CRCs.
 * 
 * @author Miles Asvachin
 */
public abstract class CRC extends Hash
{
	/**
	 * Creates a CRC, automatically choosing the most appropriate class for the given parameters.
	 */
	public static CRC createCRC(String name, int width, long polynomial, long initialValue, boolean reflectIn, boolean reflectOut, long xor)
	{
		if(width < 1)
			throw new IllegalArgumentException("CRC width must be at least 1 bit." + name);
		
		final CRC crc;

		if(width == 1) //special case
			crc = new CRC1(name);
		else if(width < 8) //Too small for a lookup-table implementation.
			crc = new SlowCRC(name, width, (int)polynomial, (int)initialValue, reflectIn, reflectOut, (int)xor);
		else if(width <= 32) //Small enough to fit in an int
			crc = new IntCRC(name, width, (int)polynomial, (int)initialValue, reflectIn, reflectOut, (int)xor);
		else if(width <= 64) //Small enough to fit in a long
			crc = new LongCRC(name, width, polynomial, initialValue, reflectIn, reflectOut, xor);
		else //Too big - try something else.
			crc = createCRC(name, width, Long.toHexString(polynomial), Long.toHexString(initialValue), reflectIn, reflectOut, Long.toHexString(xor));
			
		return crc;
	}
	
	/**
	 * Creates a CRC, automatically choosing the most appropriate class for the given parameters.
	 */
	public static CRC createCRC(String name, int width, String polynomial, String initialValue, boolean reflectIn, boolean reflectOut, String xor)
	{
		final CRC crc;
		
		BigInteger bigPoly = new BigInteger(polynomial, 16);
		BigInteger bigInit = new BigInteger(initialValue, 16);
		BigInteger bigXor = new BigInteger(xor, 16);

		if(width <= 64) //Small enough to fit in a primitive.
			crc = createCRC(name, width, bigPoly.longValue(), bigInit.longValue(), reflectIn, reflectOut, bigXor.longValue());
		else
			crc = new BigCRC(name, width, bigPoly, bigInit, reflectIn, reflectOut, bigXor);
		
		return crc;
	}

	@Override
	public HashType getHashType(){ return HashType.CRC; }
}
