package eu.asvachin.hashcalculator.hashes.crc;

import eu.asvachin.hashcalculator.utilities.BitUtilities;

/**
 * General purpose CRC algorithm with lookup table.
 * Width must be at least 8 and no more than 64 bits.
 * IntCRC is faster for widths of 32 or less. Use BigCRC if width is greater than 64.
 *
 * @version 1.0 19 June 2006
 * @author Miles Asvachin
 */
public class LongCRC extends CRC
{
	//CRC variables:
	protected final int width; //Width of the algorithm.
	protected final long polynomial; //Polynomial representation.
	protected final long initialValue; //Initial value of shiftRegister.
	protected final boolean reflectIn; //Reverse order of input bits.
	protected final boolean reflectOut; //Reverse order of output bits.
	protected final long finalXORValue; //XOR the output with this.
	
	//Variables used for calculation:
	protected long msb;
	protected long shiftRegister;
	
	private final long[] lookupTable = new long[256];
	
	protected LongCRC(String name, int w, long poly, long init, boolean refIn, boolean refOut, long xorVal)
	{
		hashName = name;
		width = w;
		polynomial = refIn ? BitUtilities.reflect(poly, w) : poly;
		initialValue = init;
		reflectIn = refIn;
		reflectOut = refIn ^ refOut;
		finalXORValue = xorVal;
		
		msb = 1L << (width - 1);
		shiftRegister = initialValue;

		calcTable();
	}
	
	/**
	 * Precalculate values to XOR with register for every possible input byte.
	 */
	private void calcTable()
	{
		if(reflectIn)
		{
			for(int i = 0; i < lookupTable.length; i++)
			{
				lookupTable[i] = i;
				
				for(int j = 0; j < 8; j++)
				{
					if((lookupTable[i] & 0x1) != 0)
					{
						lookupTable[i] = (lookupTable[i] >>> 1) ^ polynomial;
					}
					else
					{
						lookupTable[i] >>>= 1;
					}
				}
			}
		}
		else
		{
			for(int i = 0; i < lookupTable.length; i++)
			{
				lookupTable[i] = (long)i << (width - 8);

				for(int j = 0; j < 8; j++)
				{
					if((lookupTable[i] & msb) != 0)
					{
						lookupTable[i] = (lookupTable[i] << 1) ^ polynomial;
					}
					else
					{
						lookupTable[i] <<= 1;
					}
				}
			}
		}
	}
	
	@Override
	public void reset(){shiftRegister = initialValue;}
	
	@Override
	public void update(byte[] newData)
	{
		for(final byte b : newData)
		{
			//n.b, the reflected algorithm reflects both input *and* output, so may need to reflect output again.
			if(reflectIn)
			{
				shiftRegister = lookupTable[(int)((shiftRegister ^ b) & 0xFF)] ^ (shiftRegister >>> 8);
			}
			else
			{
				shiftRegister = lookupTable[(int)(((shiftRegister >> (width - 8)) ^ b) & 0xFF)] ^ (shiftRegister << 8);
			}
		}
	}
	
	@Override
	public void finish()
	{
		//Reverse bit order.
		if(reflectOut)
		{
			shiftRegister = BitUtilities.reflect(shiftRegister, width);
		}
		
		//XOR the output before returning it.
		if(finalXORValue != 0)
		{
			shiftRegister ^= finalXORValue;
		}
		
		String digestFormat = "%" + "0" + (int)Math.ceil(width/4f) + "x";
		long hashMask = 0x1;
		
		for(int i = 1; i < width; i++)
		{
			hashMask = (hashMask << 1) | 0x1;
		}
		
		digest = String.format(digestFormat, shiftRegister & hashMask);
		reset();
	}
}