package eu.asvachin.hashcalculator.hashes.crc;

import eu.asvachin.hashcalculator.utilities.BitUtilities;

/**
 * Simple CRC algorithm from http://en.wikipedia.org/wiki/Cyclic_redundancy_check
 * with some extra bits tacked on.
 * Can be configured to calculate any type of CRC up to 32 bits wide.
 * Not as fast as the lookup-table implementations.
 *
 * @version 1.0 19 June 2006
 * @author Miles Asvachin
 */
public class SlowCRC extends CRC
{
	//CRC variables:
	protected final int width; //Width of the algorithm.
	protected final int polynomial; //Polynomial representation.
	protected final int initialValue; //Initial value of shiftRegister.
	protected final boolean reflectIn; //Reverse order of input bits.
	protected final boolean reflectOut; //Reverse order of output bits.
	protected final int finalXORValue; //XOR the output with this.
	
	//Variables used for calculation:
	protected int msb;
	protected int shiftRegister;
	
	protected SlowCRC(String name, int w, int poly, int init, boolean refIn, boolean refOut, int xorVal)
	{
		hashName = name;
		width = w;
		polynomial = poly;
		initialValue = init;
		reflectIn = refIn;
		reflectOut = refOut;
		finalXORValue = xorVal;
		
		msb = 1 << (width - 1);
		shiftRegister = initialValue;
	}
	
	@Override
	public void reset(){shiftRegister = initialValue;}
	
	@Override
	public void update(byte[] newData)
	{
		for(final byte b : newData)
		{
			for(int i = 0; i < 8; i++)
			{
				if(reflectIn)
				{
					if(((shiftRegister & msb) != 0) ^ ((b & (0x1 << i))) != 0)
					{
						shiftRegister = (shiftRegister << 1) ^ polynomial;
					}
					else
					{
						shiftRegister <<= 1;
					}
				}
				else
				{
					if(((shiftRegister & msb) != 0) ^ ((b & (0x80 >> i))) != 0)
					{
						shiftRegister = (shiftRegister << 1) ^ polynomial;
					}
					else
					{
						shiftRegister <<= 1;
					}
				}
			}
		}
	}
	
	@Override
	public void finish()
	{
		//Reverse bit order.
		if(reflectOut)
		{
			shiftRegister = BitUtilities.reflect(shiftRegister, width);
		}
		
		//XOR the output before returning it.
		if(finalXORValue != 0)
		{
			shiftRegister ^= finalXORValue;
		}
		
		String digestFormat = "%" + "0" + (int)Math.ceil(width/4f) + "x";
		int hashMask = 0x1;
		
		for(int i = 1; i < width; i++)
		{
			hashMask = (hashMask << 1) | 0x1;
		}
		
		digest = String.format(digestFormat, shiftRegister & hashMask);
		reset();
	}
}
