package eu.asvachin.hashcalculator.utilities;

import java.math.BigInteger;

/**
 * Some useful functions to work on binary numbers.
 *
 * @author Miles Asvachin
 */
public final class BitUtilities
{
	private BitUtilities(){}

	/**
	 * Reverses the order of the specified bits. All other bits set to zero.
	 * 
	 * @param original The number to be reflected.
	 * @param width The number of bits to work on, starting from LSB.
	 */
	static public int reflect(int original, int width)
	{
		if(width < 1 || width > 32)
			return original;
		
		int reflected = 0;
		
		for(int i = 0; i < width ; i++)
		{
			reflected <<= 1;
			reflected |= original & 0x1;
			original >>>= 1;
		}
		
		return reflected;
	}
	
	static public long reflect(long original, int width)
	{
		if(width < 1 || width > 64)
			return original;
		
		long reflected = 0L;
		
		for(int i = 0; i < width ; i++)
		{
			reflected <<= 1;
			reflected |= original & 0x1L;
			original >>>= 1;
		}
		
		return reflected;
	}
	
	static public BigInteger reflect(BigInteger original, int width)
	{
		if(width <= 1)
			return original;
		
		BigInteger reflected = new BigInteger("0");
		
		for(int i = 0; i < width ; i++)
		{
			reflected = reflected.shiftLeft(1);
			reflected = reflected.or(original.and(BigInteger.ONE));
			original = original.shiftRight(1);
		}
		
		return reflected;
	}
	
	/**
	 * Chop off the topmost bits of a BigInteger, leaving only [widthBytes] bytes.
	 */
	static public BigInteger truncate(BigInteger original, int widthBytes)
	{
		if(original == null || widthBytes < 1)
			return original;
		
		byte[] oldBytes = original.toByteArray();
		
		if(oldBytes.length <= widthBytes)
			return new BigInteger(oldBytes);
		
		//Array is one bigger than it needs to be, otherwise it may be
		//interpreted as a negative number by the BigInteger constructor.
		byte[] newBytes = new byte[widthBytes + 1];
		System.arraycopy(oldBytes, oldBytes.length - widthBytes, newBytes, 1, widthBytes);
		
		return new BigInteger(newBytes);
	}
}